# pyftdi - A pure Python FTDI driver
# Copyright (C) 2010-2016 Emmanuel Blot <emmanuel.blot@free.fr>
# Copyright (c) 2016, Emmanuel Bouaziz <ebouaziz@free.fr>
#   Originally based on the C libftdi project
#   http://www.intra2net.com/en/developer/libftdi/
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""pyftdi - A pure Python FTDI driver on top of pyusb

Author:  Emmanuel Blot <emmanuel.blot@free.fr>
License: LGPL, originally based on libftdi C library
Caveats: Only tested with FT2232 and FT4232 FTDI devices
Require: pyusb
"""

from array import array as Array
import struct
import usb.core
import usb.util
from six import PY3
from six.moves import range
from pyftdi.usbtools import UsbTools


__all__ = ['Ftdi', 'FtdiError']


class FtdiError(IOError):
    """Communication error with the FTDI device"""


class Ftdi(object):
    """FTDI device driver"""

    # Commands
    WRITE_BYTES_PVE_MSB = 0x10
    WRITE_BYTES_NVE_MSB = 0x11
    WRITE_BITS_PVE_MSB = 0x12
    WRITE_BITS_NVE_MSB = 0x13
    WRITE_BYTES_PVE_LSB = 0x18
    WRITE_BYTES_NVE_LSB = 0x19
    WRITE_BITS_PVE_LSB = 0x1a
    WRITE_BITS_NVE_LSB = 0x1b
    READ_BYTES_PVE_MSB = 0x20
    READ_BYTES_NVE_MSB = 0x24
    READ_BITS_PVE_MSB = 0x22
    READ_BITS_NVE_MSB = 0x26
    READ_BYTES_PVE_LSB = 0x28
    READ_BYTES_NVE_LSB = 0x2c
    READ_BITS_PVE_LSB = 0x2a
    READ_BITS_NVE_LSB = 0x2e
    RW_BYTES_PVE_PVE_LSB = 0x38
    RW_BYTES_PVE_NVE_LSB = 0x39
    RW_BYTES_NVE_PVE_LSB = 0x3c
    RW_BYTES_NVE_NVE_LSB = 0x3d
    RW_BITS_PVE_PVE_LSB = 0x3a
    RW_BITS_PVE_NVE_LSB = 0x3b
    RW_BITS_NVE_PVE_LSB = 0x3e
    RW_BITS_NVE_NVE_LSB = 0x3f
    WRITE_BITS_TMS_PVE = 0x4a
    WRITE_BITS_TMS_NVE = 0x4b
    RW_BITS_TMS_PVE_PVE = 0x6a
    RW_BITS_TMS_NVE_PVE = 0x6b
    RW_BITS_TMS_PVE_NVE = 0x6e
    RW_BITS_TMS_NVE_NVE = 0x6f
    SEND_IMMEDIATE = 0x87
    WAIT_ON_HIGH = 0x88
    WAIT_ON_LOW = 0x89
    DISABLE_CLK_DIV5 = 0x8a
    ENABLE_CLK_DIV5 = 0x8b
    READ_SHORT = 0x90
    READ_EXTENDED = 0x91
    WRITE_SHORT = 0x92
    WRITE_EXTENDED = 0x93

    # Modem status
    MODEM_CTS = (1 << 4)    # Clear to send
    MODEM_DSR = (1 << 5)    # Data set ready
    MODEM_RI = (1 << 6)     # Ring indicator
    MODEM_RLSD = (1 << 7)   # Carrier detect
    MODEM_DR = (1 << 8)     # Data ready
    MODEM_OE = (1 << 9)     # Overrun error
    MODEM_PE = (1 << 10)    # Parity error
    MODEM_FE = (1 << 11)    # Framing error
    MODEM_BI = (1 << 12)    # Break interrupt
    MODEM_THRE = (1 << 13)  # Transmitter holding register
    MODEM_TEMT = (1 << 14)  # Transmitter empty
    MODEM_RCVE = (1 << 15)  # Error in RCVR FIFO

    # FTDI MPSSE commands
    SET_BITS_LOW = 0x80     # Change LSB GPIO output
    SET_BITS_HIGH = 0x82    # Change MSB GPIO output
    GET_BITS_LOW = 0x81     # Get LSB GPIO output
    GET_BITS_HIGH = 0x83    # Get MSB GPIO output
    LOOPBACK_START = 0x84   # Enable loopback
    LOOPBACK_END = 0x85     # Disable loopback
    TCK_DIVISOR = 0x86

    BITMODE_RESET = 0x00    # switch off bitbang mode
    BITMODE_BITBANG = 0x01  # classical asynchronous bitbang mode
    BITMODE_MPSSE = 0x02    # MPSSE mode, available on 2232x chips
    BITMODE_SYNCBB = 0x04   # synchronous bitbang mode
    BITMODE_MCU = 0x08      # MCU Host Bus Emulation mode,
    BITMODE_OPTO = 0x10     # Fast Opto-Isolated Serial Interface Mode
    BITMODE_CBUS = 0x20     # Bitbang on CBUS pins of R-type chips
    BITMODE_SYNCFF = 0x40   # Single Channel Synchronous FIFO mode
    BITMODE_MASK = 0x7F     # Mask for all bitmodes

    # USB control requests
    REQ_OUT = usb.util.build_request_type(
                  usb.util.CTRL_OUT,
                  usb.util.CTRL_TYPE_VENDOR,
                  usb.util.CTRL_RECIPIENT_DEVICE)
    REQ_IN = usb.util.build_request_type(
                  usb.util.CTRL_IN,
                  usb.util.CTRL_TYPE_VENDOR,
                  usb.util.CTRL_RECIPIENT_DEVICE)

    # Requests
    SIO_RESET = 0               # Reset the port
    SIO_SET_MODEM_CTRL = 1      # Set the modem control register
    SIO_SET_FLOW_CTRL = 2       # Set flow control register
    SIO_SET_BAUDRATE = 3        # Set baud rate
    SIO_SET_DATA = 4            # Set the data characteristics of the port
    SIO_POLL_MODEM_STATUS = 5   # Get line status
    SIO_SET_EVENT_CHAR = 6      # Change event character
    SIO_SET_ERROR_CHAR = 7      # Change error character
    SIO_SET_LATENCY_TIMER = 9   # Change latency timer
    SIO_GET_LATENCY_TIMER = 10  # Get latency timer
    SIO_SET_BITMODE = 11        # Change bit mode
    SIO_READ_PINS = 12          # Read GPIO pin value

    # Eeprom requests
    SIO_EEPROM = 0x90
    SIO_READ_EEPROM = SIO_EEPROM + 0   # Read EEPROM content
    SIO_WRITE_EEPROM = SIO_EEPROM + 1  # Write EEPROM content
    SIO_ERASE_EEPROM = SIO_EEPROM + 2  # Erase EEPROM content

    # Reset commands
    SIO_RESET_SIO = 0          # Reset device
    SIO_RESET_PURGE_RX = 1     # Drain RX buffer
    SIO_RESET_PURGE_TX = 2     # Drain TX buffer

    # Flow control
    SIO_DISABLE_FLOW_CTRL = 0x0
    SIO_RTS_CTS_HS = (0x1 << 8)
    SIO_DTR_DSR_HS = (0x2 << 8)
    SIO_XON_XOFF_HS = (0x4 << 8)
    SIO_SET_DTR_MASK = 0x1
    SIO_SET_DTR_HIGH = (SIO_SET_DTR_MASK | (SIO_SET_DTR_MASK << 8))
    SIO_SET_DTR_LOW = (0x0 | (SIO_SET_DTR_MASK << 8))
    SIO_SET_RTS_MASK = 0x2
    SIO_SET_RTS_HIGH = (SIO_SET_RTS_MASK | (SIO_SET_RTS_MASK << 8))
    SIO_SET_RTS_LOW = (0x0 | (SIO_SET_RTS_MASK << 8))

    # Parity bits
    PARITY_NONE, PARITY_ODD, PARITY_EVEN, PARITY_MARK, PARITY_SPACE = range(5)
    # Number of stop bits
    STOP_BIT_1, STOP_BIT_15, STOP_BIT_2 = range(3)
    # Number of bits
    BITS_7, BITS_8 = [7+i for i in range(2)]
    # Break type
    BREAK_OFF, BREAK_ON = range(2)

    # cts:  Clear to send
    # dsr:  Data set ready
    # ri:   Ring indicator
    # dcd:  Data carrier detect
    # dr:   Data ready
    # oe:   Overrun error
    # pe:   Parity error
    # fe:   Framing error
    # bi:   Break interrupt
    # thre: Transmitter holding register
    # temt: Transmitter empty
    # err:  Error in RCVR FIFO
    MODEM_STATUS = [('_0 _1 _2 _3 cts dsr ri dcd'.split()),
                    ('dr oe pe fe bi thre temt error'.split())]

    # Clocks and baudrates
    BUS_CLOCK_BASE = 6.0E6  # 6 MHz
    BUS_CLOCK_HIGH = 30.0E6  # 30 MHz
    BAUDRATE_REF_BASE = int(3.0E6)  # 3 MHz
    BAUDRATE_REF_HIGH = int(12.0E6)  # 12 MHz
    BAUDRATE_REF_SPECIAL = int(2.0E6)  # 3 MHz
    BAUDRATE_TOLERANCE = 3.0  # acceptable clock drift, in %
    BITBANG_CLOCK_MULTIPLIER = 4

    # Latency
    LATENCY_MIN = 1
    LATENCY_MAX = 255
    LATENCY_THRESHOLD = 1000

    # Special devices
    LEGACY_DEVICES = ('ft232am', )
    EXSPEED_DEVICES = ('ft2232d', )
    HISPEED_DEVICES = ('ft232h', 'ft2232h', 'ft4232h')

    def __init__(self):
        self.usb_dev = None
        self.usb_read_timeout = 5000
        self.usb_write_timeout = 5000
        self.baudrate = -1
        self.readbuffer = Array('B')
        self.readoffset = 0
        self.readbuffer_chunksize = 4 << 10  # 4KiB
        self.writebuffer_chunksize = 4 << 10  # 4KiB
        self.max_packet_size = 0
        self.interface = None
        self.index = None
        self.in_ep = None
        self.out_ep = None
        self.bitbang_mode = Ftdi.BITMODE_RESET
        self.latency = 0
        self.latency_count = 0
        self.latency_min = self.LATENCY_MIN
        self.latency_max = self.LATENCY_MAX
        self.latency_threshold = None  # disable dynamic latency
        self._wrap_api()

    # --- Public API -------------------------------------------------------

    @staticmethod
    def find_all(vps, nocache=False):
        """Find all devices that match the vendor/product pairs of the vps
           list."""
        return UsbTools.find_all(vps, nocache)

    def open(self, vendor, product, interface, index=0, serial=None,
             description=None):
        """Open a new interface to the specified FTDI device"""
        self.usb_dev = UsbTools.get_device(vendor, product, index, serial,
                                           description)
        try:
            self.usb_dev.set_configuration()
        except usb.core.USBError:
            pass
        # detect invalid interface as early as possible
        config = self.usb_dev.get_active_configuration()
        if interface > config.bNumInterfaces:
            raise FtdiError('No such FTDI port: %d' % interface)
        self._set_interface(config, interface)
        self.max_packet_size = self._get_max_packet_size()
        self._reset_device()
        self.set_latency_timer(self.LATENCY_MIN)

    def close(self):
        """Close the FTDI interface"""
        self.set_latency_timer(self.LATENCY_MAX)
        UsbTools.release_device(self.usb_dev)

    def open_mpsse(self, vendor, product, interface=1,
                   index=0, serial=None, description=None,
                   direction=0x0, initial=0x0, frequency=6.0E6, latency=16):
        """Configure the interface for MPSSE mode"""
        # Open an FTDI interface
        self.open(vendor, product, interface, index, serial, description)
        # Set latency timer
        self.set_latency_timer(latency)
        # Set chunk size
        self.write_data_set_chunksize(512)
        self.read_data_set_chunksize(512)
        # Drain input buffer
        self.purge_buffers()
        # Enable MPSSE mode
        self.set_bitmode(direction, Ftdi.BITMODE_MPSSE)
        # Configure clock
        frequency = self._set_frequency(frequency)
        # Configure I/O
        self.write_data(Array('B', (Ftdi.SET_BITS_LOW, initial, direction)))
        # Disable loopback
        self.write_data(Array('B', (Ftdi.LOOPBACK_END,)))
        self.validate_mpsse()
        # Drain input buffer
        self.purge_buffers()
        # Return the actual frequency
        return frequency

    def open_bitbang(self, vendor, product, interface=1,
                     index=0, serial=None, description=None,
                     direction=0x0, baudrate=115200, latency=16):
        """Configure the interface for BITBANG mode"""
        # Open an FTDI interface
        self.open(vendor, product, interface, index, serial, description)
        # Set latency timer
        self.set_latency_timer(latency)
        # Set chunk size
        self.write_data_set_chunksize(512)
        self.read_data_set_chunksize(512)
        # Disable loopback
        self.write_data(Array('B', (Ftdi.LOOPBACK_END,)))
        # Enable BITBANG mode
        self.set_bitmode(direction, Ftdi.BITMODE_BITBANG)
        # Configure baudrate
        self.set_baudrate(baudrate)
        self.validate_mpsse()
        # Drain input buffer
        self.purge_buffers()

    @property
    def ic_name(self):
        """Return the current type of the FTDI device as a string
           see also http://www.ftdichip.com/Support/
           Documents/TechnicalNotes/TN_100_USB_VID-PID_Guidelines.pdf
        """
        types = {0x0200: 'ft232am',
                 0x0400: 'ft232bm',
                 0x0500: 'ft2232d',
                 0x0600: 'ft232r',
                 0x0700: 'ft2232h',
                 0x0800: 'ft4232h',
                 0x0900: 'ft232h',
                 0x1000: 'ft230x'}
        return types[self.usb_dev.bcdDevice]

    @property
    def bitbang_enabled(self):
        """Tell whether some bitbang mode is activated"""
        return self.bitbang_mode not in [
                Ftdi.BITMODE_RESET,
                Ftdi.BITMODE_CBUS  # CBUS mode does not change base frequency
        ]

    @property
    def frequency_max(self):
        """Tells the maximum frequency for MPSSE clock"""
        if self.ic_name in self.HISPEED_DEVICES:
            return Ftdi.BUS_CLOCK_HIGH
        return Ftdi.BUS_CLOCK_BASE

    @property
    def fifo_sizes(self):
        """Return the (TX, RX) tupple of hardware FIFO sizes"""
        # Note that the FTDI datasheets contradict themselves, so
        # the following values may not be the right ones...
        # Note that 'TX' and 'RX' are inverted with the datasheet terminology:
        # Values here are seen from the host perspective, whereas datasheet
        # values are defined from the device perspective
        sizes = {0x0500: (384, 128),    # TX: 384, RX: 128
                 0x0600: (128, 256),    # TX: 128, RX: 256
                 0x0700: (4096, 4096),  # TX: 4KiB, RX: 4KiB
                 0x0800: (2048, 2048),  # TX: 2KiB, RX: 2KiB
                 0x0900: (1024, 1024),  # TX: 1KiB, RX: 1KiB
                 0x1000: (512, 512)}    # TX: 512, RX: 512
        return sizes.get(self.usb_dev.bcdDevice, (128, 128))  # default sizes

    def set_baudrate(self, baudrate):
        """Change the current interface baudrate"""
        if self.bitbang_enabled:
            baudrate *= Ftdi.BITBANG_CLOCK_MULTIPLIER
        actual, value, index = self._convert_baudrate(baudrate)
        delta = 100*abs(float(actual-baudrate))/baudrate
        if delta > Ftdi.BAUDRATE_TOLERANCE:
            raise ValueError('Baudrate tolerance exceeded: %.02f%% '
                             '(wanted %d, achievable %d)' %
                             (delta, baudrate, actual))
        try:
            if self.usb_dev.ctrl_transfer(Ftdi.REQ_OUT,
                                          Ftdi.SIO_SET_BAUDRATE, value,
                                          index, '', self.usb_write_timeout):
                raise FtdiError('Unable to set baudrate')
            self.baudrate = baudrate
        except usb.core.USBError as e:
            raise FtdiError('UsbError: %s' % str(e))

    def set_frequency(self, frequency):
        return self._set_frequency(frequency)

    def purge_rx_buffer(self):
        """Clear the read buffer on the chip and the internal read buffer."""
        if self._ctrl_transfer_out(Ftdi.SIO_RESET, Ftdi.SIO_RESET_PURGE_RX):
            raise FtdiError('Unable to flush RX buffer')
        # Invalidate data in the readbuffer
        self.readoffset = 0
        self.readbuffer = Array('B')

    def purge_tx_buffer(self):
        """Clear the write buffer on the chip."""
        if self._ctrl_transfer_out(Ftdi.SIO_RESET, Ftdi.SIO_RESET_PURGE_TX):
            raise FtdiError('Unable to flush TX buffer')

    def purge_buffers(self):
        """Clear the buffers on the chip and the internal read buffer."""
        self.purge_rx_buffer()
        self.purge_tx_buffer()

    # --- todo: Replace with properties -----------
    def write_data_set_chunksize(self, chunksize):
        """Configure write buffer chunk size."""
        self.writebuffer_chunksize = chunksize

    def write_data_get_chunksize(self):
        """Get write buffer chunk size."""
        return self.writebuffer_chunksize

    def read_data_set_chunksize(self, chunksize):
        """Configure read buffer chunk size."""
        # Invalidate all remaining data
        self.readoffset = 0
        self.readbuffer = Array('B')
        import sys
        if sys.platform == 'linux':
            if chunksize > 16384:
                chunksize = 16384
        self.readbuffer = []
        self.readbuffer_chunksize = chunksize

    def read_data_get_chunksize(self):
        """Get read buffer chunk size."""
        return self.readbuffer_chunksize
    # --- end of todo section ---------------------

    def set_bitmode(self, bitmask, mode):
        """Enable/disable bitbang modes."""
        value = (bitmask & 0xff) | ((mode & self.BITMODE_MASK) << 8)
        if self._ctrl_transfer_out(Ftdi.SIO_SET_BITMODE, value):
            raise FtdiError('Unable to set bitmode')
        self.bitbang_mode = mode

    def read_pins(self):
        """Directly read pin state, circumventing the read buffer.
           Useful for bitbang mode."""
        pins = self._ctrl_transfer_in(Ftdi.SIO_READ_PINS, 1)
        if not pins:
            raise FtdiError('Unable to read pins')
        return pins[0]

    def set_latency_timer(self, latency):
        """Set latency timer
           The FTDI chip keeps data in the internal buffer for a specific
           amount of time if the buffer is not full yet to decrease
           load on the usb bus."""
        if not (Ftdi.LATENCY_MIN <= latency <= Ftdi.LATENCY_MAX):
            raise ValueError("Latency out of range")
        if self._ctrl_transfer_out(Ftdi.SIO_SET_LATENCY_TIMER, latency):
            raise FtdiError('Unable to latency timer')

    def get_latency_timer(self):
        """Get latency timer"""
        latency = self._ctrl_transfer_in(Ftdi.SIO_GET_LATENCY_TIMER, 1)
        if not latency:
            raise FtdiError('Unable to get latency')
        return latency[0]

    def poll_modem_status(self):
        """Poll modem status information
           This function allows the retrieve the two status bytes of the
           device.
        """
        value = self._ctrl_transfer_in(Ftdi.SIO_POLL_MODEM_STATUS, 2)
        if not value or len(value) != 2:
            raise FtdiError('Unable to get modem status')
        status, = struct.unpack('<H', value)
        return status

    def modem_status(self):
        """Provide the current modem status as a tuple of set signals"""
        value = self._ctrl_transfer_in(Ftdi.SIO_POLL_MODEM_STATUS, 2)
        if not value or len(value) != 2:
            raise FtdiError('Unable to get modem status')
        status = []
        for pos, byte_ in enumerate(value):
            for b, v in enumerate(Ftdi.MODEM_STATUS[pos]):
                if byte_ & (1 << b):
                    status.append(Ftdi.MODEM_STATUS[pos][b])
        return tuple(status)

    def set_flowctrl(self, flowctrl):
        """Set flowcontrol for ftdi chip"""
        ctrl = {'hw': Ftdi.SIO_RTS_CTS_HS,
                'sw': Ftdi.SIO_XON_XOFF_HS,
                '': Ftdi.SIO_DISABLE_FLOW_CTRL}
        value = ctrl[flowctrl] | self.index
        try:
            if self.usb_dev.ctrl_transfer(Ftdi.REQ_OUT,
                                          Ftdi.SIO_SET_FLOW_CTRL, 0,
                                          value, '', self.usb_write_timeout):
                raise FtdiError('Unable to set flow control')
        except usb.core.USBError as e:
            raise FtdiError('UsbError: %s' % str(e))

    def set_dtr(self, state):
        """Set dtr line"""
        value = state and Ftdi.SIO_SET_DTR_HIGH or Ftdi.SIO_SET_DTR_LOW
        if self._ctrl_transfer_out(Ftdi.SIO_SET_MODEM_CTRL, value):
            raise FtdiError('Unable to set DTR line')

    def set_rts(self, state):
        """Set rts line"""
        value = state and Ftdi.SIO_SET_RTS_HIGH or Ftdi.SIO_SET_RTS_LOW
        if self._ctrl_transfer_out(Ftdi.SIO_SET_MODEM_CTRL, value):
            raise FtdiError('Unable to set RTS line')

    def set_dtr_rts(self, dtr, rts):
        """Set dtr and rts lines"""
        value = 0
        value |= dtr and Ftdi.SIO_SET_DTR_HIGH or Ftdi.SIO_SET_DTR_LOW
        value |= rts and Ftdi.SIO_SET_RTS_HIGH or Ftdi.SIO_SET_RTS_LOW
        if self._ctrl_transfer_out(Ftdi.SIO_SET_FLOW_CTRL, value):
            raise FtdiError('Unable to set DTR/RTS lines')

    def set_event_char(self, eventch, enable):
        """Set the special event character"""
        value = eventch
        if enable:
            value |= 1 << 8
        if self._ctrl_transfer_out(Ftdi.SIO_SET_EVENT_CHAR, value):
            raise FtdiError('Unable to set DTR/RTS lines')

    def set_error_char(self, errorch, enable):
        """Set error character"""
        value = errorch
        if enable:
            value |= 1 << 8
        if self._ctrl_transfer_out(Ftdi.SIO_SET_ERROR_CHAR, value):
            raise FtdiError('Unable to set DTR/RTS lines')

    def set_line_property(self, bits, stopbit, parity, break_=0):
        """Set (RS232) line characteristics"""
        bytelength = {7: Ftdi.BITS_7,
                      8: Ftdi.BITS_8}
        parities = {'N': Ftdi.PARITY_NONE,
                    'O': Ftdi.PARITY_ODD,
                    'E': Ftdi.PARITY_EVEN,
                    'M': Ftdi.PARITY_MARK,
                    'S': Ftdi.PARITY_SPACE}
        stopbits = {1: Ftdi.STOP_BIT_1,
                    1.5: Ftdi.STOP_BIT_15,
                    2: Ftdi.STOP_BIT_2}
        if parity not in parities:
            raise FtdiError("Unsupported parity")
        if bits not in bytelength:
            raise FtdiError("Unsupported byte length")
        if stopbit not in stopbits:
            raise FtdiError("Unsupported stop bits")
        value = bits & 0x0F
        try:
            value |= {Ftdi.PARITY_NONE: 0x00 << 8,
                      Ftdi.PARITY_ODD: 0x01 << 8,
                      Ftdi.PARITY_EVEN: 0x02 << 8,
                      Ftdi.PARITY_MARK: 0x03 << 8,
                      Ftdi.PARITY_SPACE: 0x04 << 8}[parities[parity]]
            value |= {Ftdi.STOP_BIT_1: 0x00 << 11,
                      Ftdi.STOP_BIT_15: 0x01 << 11,
                      Ftdi.STOP_BIT_2: 0x02 << 11}[stopbits[stopbit]]
            if break_ == Ftdi.BREAK_ON:
                value |= 0x01 << 14
        except KeyError:
            raise ValueError('Invalid line property')
        if self._ctrl_transfer_out(Ftdi.SIO_SET_DATA, value):
            raise FtdiError('Unable to set line property')

    def write_data(self, data):
        """Write data in chunks to the chip"""
        offset = 0
        size = len(data)
        try:
            while offset < size:
                write_size = self.writebuffer_chunksize
                if offset + write_size > size:
                    write_size = size - offset
                length = self._write(data[offset:offset+write_size])
                if length <= 0:
                    raise FtdiError("Usb bulk write error")
                offset += length
            return offset
        except usb.core.USBError as e:
            raise FtdiError('UsbError: %s' % str(e))

    def read_data_bytes(self, size, attempt=1):
        """Read data in chunks from the chip.
           Automatically strips the two modem status bytes transfered during
           every read."""
        # Packet size sanity check
        if not self.max_packet_size:
            raise FtdiError("max_packet_size is bogus")
        packet_size = self.max_packet_size
        length = 1  # initial condition to enter the usb_read loop
        data = Array('B')
        # everything we want is still in the cache?
        if size <= len(self.readbuffer)-self.readoffset:
            data = self.readbuffer[self.readoffset:self.readoffset+size]
            self.readoffset += size
            return data
        # something still in the cache, but not enough to satisfy 'size'?
        if len(self.readbuffer)-self.readoffset != 0:
            data = self.readbuffer[self.readoffset:]
            # end of readbuffer reached
            self.readoffset = len(self.readbuffer)
        # read from USB, filling in the local cache as it is empty
        try:
            while (len(data) < size) and (length > 0):
                while True:
                    tempbuf = self._read()
                    attempt -= 1
                    length = len(tempbuf)
                    # the received buffer contains at least one useful databyte
                    # (first 2 bytes in each packet represent the current modem
                    # status)
                    if length > 2:
                        if self.latency_threshold:
                            self.latency_count = 0
                            if self.latency != self.latency_min:
                                self.set_latency_timer(self.latency_min)
                                self.latency = self.latency_min
                        # skip the status bytes
                        chunks = (length+packet_size-1) // packet_size
                        count = packet_size - 2
                        self.readbuffer = Array('B')
                        self.readoffset = 0
                        srcoff = 2
                        for i in range(chunks):
                            self.readbuffer += tempbuf[srcoff:srcoff+count]
                            srcoff += packet_size
                        length = len(self.readbuffer)
                        break
                    else:
                        # received buffer only contains the modem status bytes
                        # no data received, may be late, try again
                        if attempt > 0:
                            continue
                        # no actual data
                        self.readbuffer = Array('B')
                        self.readoffset = 0
                        if self.latency_threshold:
                            self.latency_count += 1
                            if self.latency != self.latency_max:
                                if self.latency_count > self.latency_threshold:
                                    self.set_latency_timer(self.latency_max)
                                    self.latency = self.latency_max
                        # no more data to read?
                        return data
                if length > 0:
                    # data still fits in buf?
                    if (len(data) + length) <= size:
                        data += self.readbuffer[self.readoffset:
                                                self.readoffset+length]
                        self.readoffset += length
                        # did we read exactly the right amount of bytes?
                        if len(data) == size:
                            return data
                    else:
                        # partial copy, not enough bytes in the local cache to
                        # fulfill the request
                        part_size = min(size-len(data),
                                        len(self.readbuffer)-self.readoffset)
                        if part_size < 0:
                            raise FtdiError("Internal Error")
                        data += self.readbuffer[self.readoffset:
                                                self.readoffset+part_size]
                        self.readoffset += part_size
                        return data
        except usb.core.USBError as e:
            raise FtdiError('UsbError: %s' % str(e))
        # never reached
        raise FtdiError("Internal error")

    def read_data(self, size):
        """Read data in chunks from the chip.
           Automatically strips the two modem status bytes transfered during
           every read."""
        if PY3:
            return self.read_data_bytes(size).tobytes()
        else:
            return self.read_data_bytes(size).tostring()

    def get_cts(self):
        """Read terminal status line: Clear To Send"""
        status = self.poll_modem_status()
        return (status & self.MODEM_CTS) and True or False

    def get_dsr(self):
        """Read terminal status line: Data Set Ready"""
        status = self.poll_modem_status()
        return (status & self.MODEM_DSR) and True or False

    def get_ri(self):
        """Read terminal status line: Ring Indicator"""
        status = self.poll_modem_status()
        return (status & self.MODEM_RI) and True or False

    def get_cd(self):
        """Read terminal status line: Carrier Detect"""
        status = self.poll_modem_status()
        return (status & self.MODEM_RLSD) and True or False

    def set_dynamic_latency(self, lmin, lmax, threshold):
        """Set up or disable latency values"""
        if not threshold:
            self.latency_count = 0
            self.latency_threshold = None
        else:
            for lat in (lmin, lmax):
                if not (0 < lat < 256):
                    raise ValueError("Latency out of range: %d")
            self.latency_min = lmin
            self.latency_max = lmax
            self.latency_threshold = threshold
            self.latency = lmax
            self.set_latency_timer(self.latency)

    def validate_mpsse(self):
        # only useful in MPSSE mode
        bytes_ = self.read_data(2)
        if (len(bytes_) >= 2) and (bytes_[0] == '\xfa'):
            raise FtdiError("Invalid command @ %d" % ord(bytes_[1]))

    def get_error_string(self):
        """Wrapper for libftdi compatibility"""
        return "Unknown error"

    # --- Private implementation -------------------------------------------

    def _wrap_api(self):
        """Deal with PyUSB API breaks"""
        from six import PY3
        import inspect

        if PY3:
            args = list(inspect.signature(usb.core.Device.read).parameters)
        else:
            args, _, _, _ = \
                inspect.getargspec(usb.core.Device.read)
        if (len(args) > 2) and (args[3] == 'interface'):
            usb_api = 1  # Require "interface" parameter
        else:
            usb_api = 2
        for m in ('write', 'read'):
            setattr(self, '_%s' % m, getattr(self, '_%s_v%d' % (m, usb_api)))

    def _set_interface(self, config, ifnum):
        """Select the interface to use on the FTDI device"""
        if ifnum == 0:
            ifnum = 1
        if ifnum-1 not in range(config.bNumInterfaces):
            raise ValueError("No such interface for this device")
        self.interface = config[(ifnum-1, 0)]
        self.index = self.interface.bInterfaceNumber+1
        endpoints = sorted([ep.bEndpointAddress for ep in self.interface])
        self.in_ep, self.out_ep = endpoints[:2]

    def _reset_device(self):
        """Reset the ftdi device"""
        if self._ctrl_transfer_out(Ftdi.SIO_RESET, Ftdi.SIO_RESET_SIO):
            raise FtdiError('Unable to reset FTDI device')
        # Invalidate data in the readbuffer
        self.readoffset = 0
        self.readbuffer = Array('B')

    def _ctrl_transfer_out(self, reqtype, value, data=''):
        """Send a control message to the device"""
        try:
            return self.usb_dev.ctrl_transfer(Ftdi.REQ_OUT, reqtype, value,
                                              self.index, data,
                                              self.usb_write_timeout)
        except usb.core.USBError as e:
            raise FtdiError('UsbError: %s' % str(e))

    def _ctrl_transfer_in(self, reqtype, length):
        """Request for a control message from the device"""
        try:
            return self.usb_dev.ctrl_transfer(Ftdi.REQ_IN, reqtype, 0,
                                              self.index, length,
                                              self.usb_read_timeout)
        except usb.core.USBError as e:
            raise FtdiError('UsbError: %s' % str(e))

    def _write_v1(self, data):
        """Write to FTDI, using the deprecated API"""
        return self.usb_dev.write(self.in_ep, data,
                                  self.interface, self.usb_write_timeout)

    def _read_v1(self):
        """Read from FTDI, using the deprecated API"""
        return self.usb_dev.read(self.out_ep, self.readbuffer_chunksize,
                                 self.interface, self.usb_read_timeout)

    def _write_v2(self, data):
        """Write to FTDI, using the API introduced with pyusb 1.0.0b2"""
        return self.usb_dev.write(self.in_ep, data, self.usb_write_timeout)

    def _read_v2(self):
        """Read from FTDI, using the API introduced with pyusb 1.0.0b2"""
        return self.usb_dev.read(self.out_ep, self.readbuffer_chunksize,
                                 self.usb_read_timeout)

    def _get_max_packet_size(self):
        """Retrieve the maximum length of a data packet"""
        if not self.usb_dev:
            import errno
            raise IOError("Device is not yet known", errno.ENODEV)
        if not self.interface:
            import errno
            raise IOError("Interface is not yet known", errno.ENODEV)
        if self.ic_name in self.HISPEED_DEVICES:
            packet_size = 512
        else:
            packet_size = 64
        endpoint = self.interface[0]
        packet_size = endpoint.wMaxPacketSize
        return packet_size

    def _convert_baudrate(self, baudrate):
        """Convert a requested baudrate into the closest possible baudrate
           that can be assigned to the FTDI device"""
        if baudrate < ((2*self.BAUDRATE_REF_BASE)//(2*16384+1)):
            raise ValueError('Invalid baudrate (too low)')
        if baudrate > self.BAUDRATE_REF_BASE:
            if self.ic_name not in self.HISPEED_DEVICES or \
               baudrate > self.BAUDRATE_REF_HIGH:
                    raise ValueError('Invalid baudrate (too high)')
            refclock = self.BAUDRATE_REF_HIGH
            hispeed = True
        else:
            refclock = self.BAUDRATE_REF_BASE
            hispeed = False
        # AM legacy device only supports 3 sub-integer dividers, where the
        # other devices supports 8 sub-integer dividers
        am_adjust_up = [0, 0, 0, 1, 0, 3, 2, 1]
        am_adjust_dn = [0, 0, 0, 1, 0, 1, 2, 3]
        # Sub-divider code are not ordered in the natural order
        frac_code = [0, 3, 2, 4, 1, 5, 6, 7]
        divisor = (refclock*8) // baudrate
        if self.ic_name in self.LEGACY_DEVICES:
            # Round down to supported fraction (AM only)
            divisor -= am_adjust_dn[divisor & 7]
        # Try this divisor and the one above it (because division rounds down)
        best_divisor = 0
        best_baud = 0
        best_baud_diff = 0
        for i in range(2):
            try_divisor = divisor + i
            if not hispeed:
                # Round up to supported divisor value
                if try_divisor <= 8:
                    # Round up to minimum supported divisor
                    try_divisor = 8
                elif self.ic_name not in self.LEGACY_DEVICES and \
                        try_divisor < 12:
                    # BM doesn't support divisors 9 through 11 inclusive
                    try_divisor = 12
                elif divisor < 16:
                    # AM doesn't support divisors 9 through 15 inclusive
                    try_divisor = 16
                else:
                    if self.ic_name in self.LEGACY_DEVICES:
                        # Round up to supported fraction (AM only)
                        try_divisor += am_adjust_up[try_divisor & 7]
                        if try_divisor > 0x1FFF8:
                            # Round down to maximum supported div value (AM)
                            try_divisor = 0x1FFF8
                    else:
                        if try_divisor > 0x1FFFF:
                            # Round down to maximum supported div value (BM)
                            try_divisor = 0x1FFFF
            # Get estimated baud rate (to nearest integer)
            baud_estimate = ((refclock*8) + (try_divisor//2))//try_divisor
            # Get absolute difference from requested baud rate
            if baud_estimate < baudrate:
                baud_diff = baudrate - baud_estimate
            else:
                baud_diff = baud_estimate - baudrate
            if (i == 0) or (baud_diff < best_baud_diff):
                # Closest to requested baud rate so far
                best_divisor = try_divisor
                best_baud = baud_estimate
                best_baud_diff = baud_diff
                if baud_diff == 0:
                    break
        # Encode the best divisor value
        encoded_divisor = (best_divisor >> 3) | \
                          (frac_code[best_divisor & 7] << 14)
        # Deal with special cases for encoded value
        if encoded_divisor == 1:
            encoded_divisor = 0  # 3000000 baud
        elif encoded_divisor == 0x4001:
            encoded_divisor = 1  # 2000000 baud (BM only)
        # Split into "value" and "index" values
        value = encoded_divisor & 0xFFFF
        if self.ic_name in self.EXSPEED_DEVICES + self.HISPEED_DEVICES:
            index = (encoded_divisor >> 8) & 0xFFFF
            index &= 0xFF00
            index |= self.index
        else:
            index = (encoded_divisor >> 16) & 0xFFFF
        if hispeed:
            index |= 1 << 9  # use hispeed mode
        return (best_baud, value, index)

    def _set_frequency(self, frequency):
        """Convert a frequency value into a TCK divisor setting"""
        if frequency > self.frequency_max:
            raise FtdiError("Unsupported frequency: %f" % frequency)
        if frequency <= Ftdi.BUS_CLOCK_BASE:
            divcode = Ftdi.ENABLE_CLK_DIV5
            divisor = int(Ftdi.BUS_CLOCK_BASE/frequency)-1
            actual_freq = Ftdi.BUS_CLOCK_BASE/(divisor+1)
        elif frequency <= Ftdi.BUS_CLOCK_HIGH:
            # not supported on non-H device, however it seems that 2232D
            # devices simply ignore the settings. Could be improved though
            divcode = Ftdi.DISABLE_CLK_DIV5
            divisor = int(Ftdi.BUS_CLOCK_HIGH/frequency)-1
            actual_freq = Ftdi.BUS_CLOCK_HIGH/(divisor+1)
        else:
            raise FtdiError("Unsupported frequency: %f" % frequency)
        # FTDI expects little endian
        if self.ic_name in self.HISPEED_DEVICES:
            cmd = Array('B', (divcode,))
        else:
            cmd = Array('B')
        cmd.extend((Ftdi.TCK_DIVISOR, divisor & 0xff, (divisor >> 8) & 0xff))
        self.write_data(cmd)
        self.validate_mpsse()
        # Drain input buffer
        self.purge_rx_buffer()
        return actual_freq

    def __get_timeouts(self):
        return self.usb_read_timeout, self.usb_write_timeout

    def __set_timeouts(self, timeouts):
        (read_timeout, write_timeout) = timeouts
        self.usb_read_timeout = read_timeout
        self.usb_write_timeout = write_timeout

    timeouts = property(__get_timeouts, __set_timeouts)



# Copyright (c) 2014-2016, Emmanuel Blot <emmanuel.blot@free.fr>
# Copyright (c) 2016, Emmanuel Bouaziz <ebouaziz@free.fr>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Neotion nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL NEOTION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pyftdi.ftdi import Ftdi
from struct import pack as spack


__all__ = ['GpioController']


class GpioException(IOError):
    """Base class for GPIO errors"""


class GpioController(object):
    """GPIO controller for an FTDI port"""

    MASK = 0xff

    def __init__(self):
        self._ftdi = None
        self._direction = 0

    def __del__(self):
        self.close()

    @property
    def direction(self):
        return self._direction

    @property
    def is_connected(self):
        return bool(self._ftdi)

    def open(self, vendor, product, interface, direction, **kwargs):
        """
        """
        for k in ('direction',):
            if k in kwargs:
                del kwargs[k]
        try:
            ftdi = Ftdi()
            ftdi.open_bitbang(vendor, product, interface, direction=direction,
                              **kwargs)
            self._ftdi = ftdi
        except IOError as e:
            raise GpioException('Unable to open USB port: %s' % str(e))
        self._direction = direction

    def close(self):
        """Close the FTDI interface"""
        if self._ftdi:
            self._ftdi.close()
            self._ftdi = None

    def set_direction(self, direction):
        """Update the GPIO pin direction.

           :param direction: a bitfield of GPIO pins. Each bit represent a GPIO
                             pin, where '1' sets the pin as output and '0' sets
                             the pin as input/high-Z.
        """
        if direction > self.MASK:
            raise GpioException("Invalid direction mask")
        self._direction = direction
        self._ftdi.set_bitmode(self.direction, Ftdi.BITMODE_BITBANG)

    def read_port(self):
        """Read the GPIO input pin electrical level.

           :param value: a bitfield of GPIO pins. Each bit represent a GPIO
                         pin, where '1' sets the pin to high level and '0'
                         sets the pin to low level. GPIO pins that are
                         configured as Input are left unaltered.
        """
        if not self.is_connected:
            raise GpioException('Not connected')
        return self._ftdi.read_pins()

    def write_port(self, value):
        """Set the GPIO output pin electrical level.

           :param value: a bitfield of GPIO pins. Each bit represent a GPIO
                         pin, where '1' reports a high level on the matching
                         pin and '0' reports a low level. GPIO pins that are
                         configured as Output should be ignored.
        """
        if not self.is_connected:
            raise GpioException('Not connected')
        if value > self._direction or (value & ~self._direction):
            raise GpioException("Invalid value")
        self._ftdi.write_data(spack('<B', value))



# Copyright (c) 2008-2012, Neotion
# Copyright (c) 2011-2016, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Neotion nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL NEOTION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import time
from pyftdi.usbtools import UsbTools, UsbToolsError
from serial import SerialBase, VERSION as pyserialver
from six.moves import range

__all__ = ['UsbSerial']


class UsbSerial(SerialBase):
    """Base class for Serial port implementation compatible with pyserial API
       using a USB device.
    """

    BAUDRATES = sorted([9600 * (x+1) for x in range(6)] +
                       list(range(115200, 1000000, 115200)) +
                       list(range(1000000, 13000000, 100000)))

    PYSERIAL_VERSION = tuple([int(x) for x in pyserialver.split('.')])

    def makeDeviceName(self, port):
        return port

    def open(self, devclass, scheme, vdict, pdict, default_vendor):
        """Open the initialized serial port"""
        from serial import SerialException
        if self._port is None:
            raise SerialException("Port must be configured before use.")
        try:
            vendor, product, interface, sernum, ix = UsbTools.parse_url(
                self.portstr, devclass, scheme, vdict, pdict, default_vendor)
        except UsbToolsError as e:
            raise SerialException(str(e))
        try:
            self.udev = devclass()
            self.udev.open(vendor, product, interface, ix, sernum)
            self.flushOutput()
            self.flushInput()
        except IOError:
            raise SerialException('Unable to open USB port %s' % self.portstr)
        self._set_open_state(True)
        self._reconfigurePort()
        self._product = product

    def close(self):
        """Close the open port"""
        self._set_open_state(False)
        self.udev.close()
        self.udev = None

    def read(self, size=1):
        """Read size bytes from the serial port. If a timeout is set it may
           return less characters as requested. With no timeout it will block
           until the requested number of bytes is read."""
        data = bytearray()
        start = time.time()
        while size > 0:
            buf = self.udev.read_data(size)
            data += buf
            size -= len(buf)
            if self._timeout > 0:
                if buf:
                    break
                ms = time.time()-start
                if ms > self._timeout:
                    break
            time.sleep(0.01)
        return data

    def write(self, data):
        """Output the given string over the serial port."""
        self.udev.write_data(data)

    def flush(self):
        """Flush of file like objects. In this case, wait until all data
           is written."""
        # do nothing

    def flushInput(self):
        """Clear input buffer, discarding all that is in the buffer."""
        self.udev.purge_rx_buffer()

    def flushOutput(self):
        """Clear output buffer, aborting the current output and
        discarding all that is in the buffer."""
        self.udev.purge_tx_buffer()

    def sendBreak(self):
        """Send break condition."""
        # Not supported
        pass

    def setRTS(self, level=True):
        """Set terminal status line: Request To Send"""
        self.udev.set_rts(level)

    def setDTR(self, level=True):
        """Set terminal status line: Data Terminal Ready"""
        self.udev.set_dtr(level)

    def getCTS(self):
        """Read terminal status line: Clear To Send"""
        return self.udev.get_cts()

    def getDSR(self):
        """Read terminal status line: Data Set Ready"""
        return self.udev.get_dsr()

    def getRI(self):
        """Read terminal status line: Ring Indicator"""
        return self.udev.get_ri()

    def getCD(self):
        """Read terminal status line: Carrier Detect"""
        return self.udev.get_cd()

    def inWaiting(self):
        """Return the number of characters currently in the input buffer."""
        # not implemented
        return 0

    @property
    def fifoSizes(self):
        """Return the (TX, RX) tupple of hardware FIFO sizes"""
        return self.udev.fifo_sizes

    def _reconfigurePort(self):
        try:
            self.udev.set_baudrate(self._baudrate)
            self.udev.set_line_property(self._bytesize,
                                        self._stopbits,
                                        self._parity)
            if self._rtscts:
                self.udev.set_flowctrl('hw')
            elif self._xonxoff:
                self.udev.set_flowctrl('sw')
            else:
                self.udev.set_flowctrl('')
            try:
                self.udev.set_dynamic_latency(2, 200, 400)
            except AttributeError:
                # backend does not support this feature
                pass
        except IOError as e:
            from serial import SerialException
            err = self.udev.get_error_string()
            raise SerialException("%s (%s)" % (str(e), err))

    _reconfigure_port = _reconfigurePort

    def _set_open_state(self, open_):
        if self.PYSERIAL_VERSION < (3, 0):
            self._isOpen = bool(open_)
        else:
            self.is_open = bool(open_)



# Copyright (c) 2008-2016, Neotion
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Neotion nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL NEOTION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from io import RawIOBase

from pyftdi.ftdi import Ftdi, FtdiError
from pyftdi.serialext.serialusb import UsbSerial

__all__ = ['Serial']


class FtdiSerial(UsbSerial):
    """Serial port implementation for FTDI compatible with pyserial API"""

    BACKEND = 'pyftdi'
    SCHEME = 'ftdi'
    FTDI_VENDOR = 0x403
    VENDOR_IDS = {'ftdi': FTDI_VENDOR}
    PRODUCT_IDS = {
        FTDI_VENDOR:
            {'232': 0x6001,
             '232r': 0x6001,
             '232h': 0x6014,
             '2232': 0x6010,
             '4232': 0x6011,
             '230x': 0x6015,
             'ft232': 0x6001,
             'ft232r': 0x6001,
             'ft232h': 0x6014,
             'ft2232': 0x6010,
             'ft4232': 0x6011,
             'ft230x': 0x6015
             }
        }
    DEFAULT_VENDOR = FTDI_VENDOR

    def open(self):
        """Open the initialized serial port"""
        from serial.serialutil import SerialException
        try:
            UsbSerial.open(self, Ftdi,
                           FtdiSerial.SCHEME,
                           FtdiSerial.VENDOR_IDS,
                           FtdiSerial.PRODUCT_IDS,
                           FtdiSerial.DEFAULT_VENDOR)
        except FtdiError as e:
            raise SerialException(str(e))

    @classmethod
    def add_custom_vendor(cls, vid, vidname=''):
        """Add a custom USB vendor identifier.
           It can be useful to use a pretty URL for opening FTDI device
        """
        if vid in cls.VENDOR_IDS:
            raise ValueError('Vendor ID 0x%04x already registered' % vid)
        if not vidname:
            vidname = '0x%04x' % vid
        cls.VENDOR_IDS[vidname] = vid

    @classmethod
    def add_custom_product(cls, vid, pid, pidname=''):
        """Add a custom USB product identifier.
           It is required for opening FTDI device with non-standard VID/PID
           USB identifiers.
        """
        if vid not in cls.PRODUCT_IDS:
            cls.PRODUCT_IDS[vid] = {}
        elif pid in cls.PRODUCT_IDS[vid]:
            raise ValueError('Product ID 0x%04x already registered' % vid)
        if not pidname:
            pidname = '0x%04x' % pid
        cls.PRODUCT_IDS[vid][pidname] = pid


# assemble Serial class with the platform specific implementation and the base
# for file-like behavior.
class Serial(FtdiSerial, RawIOBase):
    pass



# Copyright (c) 2010-2016, Emmanuel Blot <emmanuel.blot@free.fr>
# Copyright (c) 2016, Emmanuel Bouaziz <ebouaziz@free.fr>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Neotion nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL NEOTION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import struct
from array import array as Array
from pyftdi.ftdi import Ftdi
from six import PY3

__all__ = ['SpiPort', 'SpiController']


class SpiIOError(IOError):
    """SPI I/O error"""


class SpiPort(object):
    """SPI port

       An SPI port is never instanciated directly.

       Use SpiController.get_port() method to obtain an SPI port

       :Example:

            ctrl = SpiController(silent_clock=False)
            ctrl.configure(0x1234, 0x5678, 1)
            spi = ctrl.get_port(1)
            spi.set_frequency(1000000)
            # send 2 bytes
            spi.exchange([0x12, 0x34])
            # send 2 bytes, then receive 2 bytes
            out = spi.exchange([0x12, 0x34], 2)
            # send 2 bytes, then receive 4 bytes, manage the transaction
            out = spi.exchange([0x12, 0x34], 2, True, False)
            out.extend(spi.exchange([], 2, False, True))
    """

    def __init__(self, controller, cs_cmd):
        self._controller = controller
        self._cs_cmd = cs_cmd
        self._frequency = self._controller.frequency

    def exchange(self, out='', readlen=0, start=True, stop=True):
        """Perform an exchange or a transaction with the SPI slave

           .. note:: Exchange is a dual half-duplex transmission: output bytes
                     are sent to the slave, then bytes are received from the
                     slave. It is not possible to perform a full duplex
                     exchange for now, although this feature could be easily
                     implemented.

           :param out: an array of bytes to send to the SPI slave,
                       may be empty to only read out data from the slave
           :param readlen: count of bytes to read out from the slave,
                       may be zero to only write to the slave
           :param start: whether to start an SPI transaction, i.e. activate
                         the /CS line for the slave. Use False to resume a
                         previously started transaction
           :param stop: whether to desactivete the /CS line for the slave. Use
                       False if the transaction should complete with a further
                       call to exchange()
           :return: an array of bytes containing the data read out from the
                    slave
        """
        return self._controller._exchange(self._frequency, out, readlen,
                                          start and self._cs_cmd, stop)

    def read(self, readlen=0, start=True, stop=True):
        """Read out bytes from the slave"""
        return self._controller._exchange(self._frequency, [], readlen,
                                          start and self._cs_cmd, stop)

    def write(self, out, start=True, stop=True):
        """Write bytes to the slave"""
        return self._controller._exchange(self._frequency, out, 0,
                                          start and self._cs_cmd, stop)

    def flush(self):
        """Force the flush of the HW FIFOs"""
        self._controller._flush()

    def set_frequency(self, frequency):
        """Change SPI bus frequency"""
        self._frequency = min(frequency, self._controller.frequency_max)

    @property
    def frequency(self):
        """Return the current SPI bus block"""
        return self._frequency


class SpiController(object):
    """SPI master.

        :param silent_clock: should be set to avoid clocking out SCLK when all
                             /CS signals are released. This clock beat is used
                             to enforce a delay between /CS signal activation.

                             When weird SPI devices are used, SCLK beating may
                             cause trouble. In this case, silent_clock should
                             be set but beware that SCLK line should be fitted
                             with a pull-down resistor, as SCLK is high-Z
                             during this short period of time.
        :param cs_count: is the number of /CS lines (one per device to drive on
                         the SPI bus)
    """

    SCK_BIT = 0x01
    DO_BIT = 0x02
    DI_BIT = 0x04
    CS_BIT = 0x08
    PAYLOAD_MAX_LENGTH = 0x10000  # 16 bits max

    def __init__(self, silent_clock=False, cs_count=4, turbo=True):
        self._ftdi = Ftdi()
        self._cs_bits = (((SpiController.CS_BIT << cs_count) - 1) &
                         ~(SpiController.CS_BIT - 1))
        self._ports = [None] * cs_count
        self._direction = (self._cs_bits |
                           SpiController.DO_BIT |
                           SpiController.SCK_BIT)
        self._turbo = turbo
        self._cs_high = Array('B')
        if self._turbo:
            if silent_clock:
                # Set SCLK as input to avoid emitting clock beats
                self._cs_high.extend((Ftdi.SET_BITS_LOW, self._cs_bits,
                                      self._direction & ~SpiController.SCK_BIT))
            # /CS to SCLK delay, use 8 clock cycles as a HW tempo
            self._cs_high.extend((Ftdi.WRITE_BITS_TMS_NVE, 8-1, 0xff))
        # Restore idle state
        self._cs_high.extend((Ftdi.SET_BITS_LOW, self._cs_bits,
                              self._direction))
        if not self._turbo:
            self._cs_high.append(Ftdi.SEND_IMMEDIATE)
        self._immediate = Array('B', (Ftdi.SEND_IMMEDIATE,))
        self._frequency = 0.0

    def configure(self, vendor, product, interface, **kwargs):
        """Configure the FTDI interface as a SPI master"""
        for k in ('direction', 'initial'):
            if k in kwargs:
                del kwargs[k]
        self._frequency = \
            self._ftdi.open_mpsse(vendor, product, interface,
                                  direction=self._direction,
                                  initial=self._cs_bits,  # /CS all high
                                  **kwargs)

    def terminate(self):
        """Close the FTDI interface"""
        if self._ftdi:
            self._ftdi.close()
            self._ftdi = None

    def get_port(self, cs):
        """Obtain a SPI port to drive a SPI device selected by cs"""
        if not self._ftdi:
            raise SpiIOError("FTDI controller not initialized")
        if cs >= len(self._ports):
            raise SpiIOError("No such SPI port")
        if not self._ports[cs]:
            cs_state = 0xFF & ~((SpiController.CS_BIT << cs) |
                                SpiController.SCK_BIT |
                                SpiController.DO_BIT)
            cs_cmd = Array('B', (Ftdi.SET_BITS_LOW,
                                 cs_state,
                                 self._direction))
            self._ports[cs] = SpiPort(self, cs_cmd)
            self._flush()
        return self._ports[cs]

    @property
    def frequency_max(self):
        """Returns the maximum SPI clock"""
        return self._ftdi.frequency_max

    @property
    def frequency(self):
        """Returns the current SPI clock"""
        return self._frequency

    def _exchange(self, frequency, out, readlen, cs_cmd=None, complete=True):
        """Perform a half-duplex exchange or transaction with the SPI slave

           :param frequency: SPI bus clock
           :param out: an array of bytes to send to the SPI slave,
                       may be empty to only read out data from the slave
           :param readlen: count of bytes to read out from the slave,
                       may be zero to only write to the slave
           :param cs_cmd: the prolog sequence to activate the /CS line on the
                       SPI bus. May be empty to resume a previously started
                       transaction
           :param complete: whether to send the epilog sequence to move the
                       /CS line back to the idle state. May be force to False
                       if another part of a transaction is expected
           :return: an array of bytes containing the data read out from the
                    slave
        """
        if not self._ftdi:
            raise SpiIOError("FTDI controller not initialized")
        if len(out) > SpiController.PAYLOAD_MAX_LENGTH:
            raise SpiIOError("Output payload is too large")
        if readlen > SpiController.PAYLOAD_MAX_LENGTH:
            raise SpiIOError("Input payload is too large")
        if self._frequency != frequency:
            self._ftdi.set_frequency(frequency)
            # store the requested value, not the actual one (best effort)
            self._frequency = frequency
        cmd = cs_cmd and Array('B', cs_cmd) or Array('B')
        writelen = len(out)
        if writelen:
            write_cmd = struct.pack('<BH', Ftdi.WRITE_BYTES_NVE_MSB,
                                    writelen-1)
            if PY3:
                cmd.frombytes(write_cmd)
            else:
                cmd.fromstring(write_cmd)
            cmd.extend(out)
        if readlen:
            read_cmd = struct.pack('<BH', Ftdi.READ_BYTES_NVE_MSB,
                                   readlen-1)
            if PY3:
                cmd.frombytes(read_cmd)
            else:
                cmd.fromstring(read_cmd)
            cmd.extend(self._immediate)
            if self._turbo:
                if complete:
                    cmd.extend(self._cs_high)
                self._ftdi.write_data(cmd)
            else:
                self._ftdi.write_data(cmd)
                if complete:
                    self._ftdi.write_data(self._cs_high)
            # USB read cycle may occur before the FTDI device has actually
            # sent the data, so try to read more than once if no data is
            # actually received
            data = self._ftdi.read_data_bytes(readlen, 4)
        elif writelen:
            if self._turbo:
                if complete:
                    cmd.extend(self._cs_high)
                self._ftdi.write_data(cmd)
            else:
                self._ftdi.write_data(cmd)
                if complete:
                    self._ftdi.write_data(self._cs_high)
            data = Array('B')
        return data

    def _flush(self):
        """Flush the HW FIFOs"""
        self._ftdi.write_data(self._immediate)
        self._ftdi.purge_buffers()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2010-2016 Emmanuel Blot <emmanuel.blot@free.fr>
# Copyright (c) 2010-2016 Neotion
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

try:
    # try to use setuptools first, so extended command set such as
    # python setup.py develop is available
    from setuptools import setup
except ImportError:
    # if setuptools package is not available, fall back to the default
    # distribution package.
    from distutils.core import setup
from pyftdi import __version__ as VERSION


def _read(fname):
    import os
    return open(os.path.join(os.path.dirname(__file__),
                'pyftdi', fname)).read()

setup(
    name='pyftdi',
    version=VERSION,
    description='FTDI device driver (pure Python)',
    author='Emmanuel Blot',
    author_email='emmanuel.blot@free.fr',
    license='LGPL v2',
    keywords='driver ftdi usb serial spi rs232 gpio bit-bang',
    url='http://github.com/eblot/pyftdi',
    download_url='https://github.com/eblot/pyftdi/archive/v%s.tar.gz' %
                 VERSION,
    packages=['pyftdi', 'pyftdi.serialext'],
    extras_require={'spiflash': []},
    package_data={'pyftdi': ['*.rst'],
                  'pyftdi.serialext': ['*.rst']},
    requires=['pyusb (>= 1.0.0b1)',
              'pyserial (>= 2.6)',
              'six'],
    install_requires=['pyusb>=1.0.0b1',
                      'pyserial>=2.6',
                      'six'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Library or '
            'Lesser General Public License (LGPL)',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Hardware :: Hardware Drivers',
    ],
    long_description=_read('README.rst'),
)



# Copyright (c) 2008-2016, Neotion
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Neotion nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL NEOTION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import types
from pyftdi.misc import hexdump
from time import time
from six import print_

__all__ = ['SerialLogger']


class SerialLogger(object):
    """Serial port wrapper to log input/output data to a log file"""

    def __init__(self, logpath):
        try:
            self._logger = open(logpath, "wt")
        except IOError as e:
            print_("Cannot log data to %s: %s" % (logpath, str(e)),
                   file=sys.stderr)
        self._port = None
        self._methods = {}
        self._last = time()

    def spy(self, port):
        self._port = port
        methods = [m for m in self.__class__.__dict__
                   if not m.startswith('_') and
                   hasattr(getattr(self.__class__, m), '__call__') and
                   m != 'spy']
        # replace the spied instance method with our own methods
        for method_name in methods:
            try:
                old_method = getattr(port.__class__, method_name)
            except AttributeError:
                old_method = None
            if old_method:
                new_method = getattr(self.__class__, method_name)
                setattr(port, method_name, types.MethodType(new_method, self))
                self._methods[method_name] = old_method

    def _print(self, header, string):
        if self._logger:
            now = time()
            delta = (now-self._last)*1000
            self._last = now
            print_("%s (%3.3f ms):\n%s" % (header, delta, string),
                   file=self._logger)
            self._logger.flush()

    def _log_read(self, data):
        try:
            self._print('READ', hexdump(data))
        except Exception as e:
            print_('Cannot log input data (%s)' % e, file=sys.stderr)

    def _log_write(self, data):
        try:
            self._print('WRITE', hexdump(data))
        except Exception as e:
            print_('Cannot log output data (%s)' % e, data, file=sys.stderr)

    def _log_flush(self, type_):
        try:
            self._print('FLUSH', type_)
        except Exception as e:
            print_('Cannot log flush action (%s)' % e, file=sys.stderr)

    def _log_waiting(self, count):
        try:
            self._print('INWAITING', '%d' % count)
        except Exception as e:
            print_('Cannot log inwaiting (%s)' % e, file=sys.stderr)

    def _log_setBaudrate(self, baudrate):
        try:
            self._print('SETBAUDRATE', '%d' % baudrate)
        except Exception as e:
            print_('Cannot log setBaudrate (%s)' % e, file=sys.stderr)

    def _log_setDTR(self, hwreset):
        try:
            self._print('SETDTR', '%s' % hwreset)
        except Exception as e:
            print_('Cannot log setDTR (%s)' % e, file=sys.stderr)

    def _log_setRTS(self, startmode):
        try:
            self._print('SETRTS', '%d' % startmode)
        except Exception as e:
            print_('Cannot log setRTS (%s)' % e, file=sys.stderr)

    def close(self):
        self._logger.close()
        self._methods['close'](self._port)

    def read(self, size=1):
        data = self._methods['read'](self._port, size)
        self._log_read(data)
        return data

    def write(self, data):
        self._methods['write'](self._port, data)
        if len(data):
            self._log_write(data)

    def inWaiting(self):
        wait = self._methods['inWaiting'](self._port)
        self._log_waiting(wait)
        return wait

    def flush(self):
        self._log_flush('I+O')
        self._methods['flush'](self._port)

    def flushInput(self):
        self._log_flush('I')
        self._methods['flushInput'](self._port)

    def flushOutput(self):
        self._log_flush('O')
        self._methods['flushOutput'](self._port)

    def setBaudrate(self, baudrate):
        self._log_setBaudrate(baudrate)
        self._methods['setBaudrate'](self._port, baudrate)

    def setDTR(self, hwreset):
        self._log_setDTR(hwreset)
        self._methods['setDTR'](self._port, hwreset)

    def setRTS(self, startmode):
        self._log_setRTS(startmode)
        self._methods['setRTS'](self._port, startmode)



# Copyright (C) 2010-2016 Emmanuel Blot <emmanuel.blot@free.fr>
# Copyright (C) 2016 Emmanuel Bouaziz <ebouaziz@free.fr>
# All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import threading
import usb.core
import usb.util
from pyftdi.misc import to_int
from six import print_
from six.moves.urllib.parse import urlsplit

__all__ = ['UsbTools']


class UsbToolsError(Exception):
    """UsbTools error"""


class UsbTools(object):
    """Helpers to obtain information about connected USB devices."""

    # Need to maintain a list of reference USB devices, to circumvent a
    # limitation in pyusb that prevents from opening several times the same
    # USB device. The following dictionary used bus/address/vendor/product keys
    # to track (device, refcount) pairs
    Devices = {}
    Lock = threading.RLock()
    UsbDevices = {}
    UsbApi = None

    @staticmethod
    def find_all(vps, nocache=False):
        """Find all devices that match the vendor/product pairs of the vps
           list."""
        devs = set()
        for v, p in vps:
            devs.update(UsbTools._find_devices(v, p, nocache))
        devices = set()
        for dev in devs:
            ifcount = max([cfg.bNumInterfaces for cfg in dev])
            sernum = UsbTools.get_string(dev, dev.iSerialNumber)
            description = UsbTools.get_string(dev, dev.iProduct)
            devices.add((dev.idVendor, dev.idProduct, sernum, ifcount,
                         description))
        return list(devices)

    @classmethod
    def flush_cache(cls):
        cls.Lock.acquire()
        cls.UsbDevices = {}
        cls.Lock.release()

    @classmethod
    def get_device(cls, vendor, product, index, serial, description):
        """Find a previously open device with the same vendor/product
           or initialize a new one, and return it"""
        cls.Lock.acquire()
        try:
            if index or serial or description:
                dev = None
                if not vendor:
                    raise ValueError('Vendor identifier is required')
                devs = cls._find_devices(vendor, product)
                if description:
                    devs = [dev for dev in devs if
                            UsbTools.get_string(dev, dev.iProduct) ==
                            description]
                if serial:
                    devs = [dev for dev in devs if
                            UsbTools.get_string(dev, dev.iSerialNumber) ==
                            serial]
                try:
                    dev = devs[index]
                except IndexError:
                    raise IOError("No such device")
            else:
                devs = cls._find_devices(vendor, product)
                dev = devs and list(devs)[0] or None
            if not dev:
                raise IOError('Device not found')
            try:
                devkey = (dev.bus, dev.address, vendor, product)
                if None in devkey[0:2]:
                    raise AttributeError('USB backend does not support bus '
                                         'enumeration')
            except AttributeError:
                devkey = (vendor, product)
            if devkey not in cls.Devices:
                for configuration in dev:
                    # we need to detach any kernel driver from the device
                    # be greedy: reclaim all device interfaces from the kernel
                    for interface in configuration:
                        ifnum = interface.bInterfaceNumber
                        try:
                            if not dev.is_kernel_driver_active(ifnum):
                                continue
                            dev.detach_kernel_driver(ifnum)
                        except NotImplementedError:
                            # only libusb 1.x backend implements this method
                            break
                        except usb.core.USBError:
                            pass
                # only change the active configuration if the active one is
                # not the first. This allows other libusb sessions running
                # with the same device to run seamlessly.
                if dev.get_active_configuration().bConfigurationValue != 1:
                    try:
                        dev.set_configuration()
                    except usb.core.USBError:
                        pass
                cls.Devices[devkey] = [dev, 1]
            else:
                cls.Devices[devkey][1] += 1
            return cls.Devices[devkey][0]
        finally:
            cls.Lock.release()

    @classmethod
    def release_device(cls, usb_dev):
        """Release a previously open device, if it not used anymore"""
        # Lookup for ourselves in the class dictionary
        cls.Lock.acquire()
        try:
            for devkey in cls.Devices:
                dev, refcount = cls.Devices[devkey]
                if dev == usb_dev:
                    # found
                    if refcount > 1:
                        # another interface is open, decrement
                        cls.Devices[devkey][1] -= 1
                    else:
                        # last interface in use, release
                        usb.util.dispose_resources(cls.Devices[devkey][0])
                        del cls.Devices[devkey]
                    break
        finally:
            cls.Lock.release()

    @classmethod
    def _find_devices(cls, vendor, product, nocache=False):
        """Find an USB device and return it.
           This code re-implements the usb.core.find() method using a local
           cache to avoid calling several times the underlying LibUSB and the
           system USB calls to enumerate the available USB devices. As these
           calls are time-hungry (about 1 second/call), the enumerated devices
           are cached. It consumes a bit more memory but dramatically improves
           start-up time.
           Hopefully, this kludge is temporary and replaced with a better
           implementation from PyUSB at some point.
        """
        cls.Lock.acquire()
        try:
            backend = None
            candidates = ('libusb1', 'libusb10', 'libusb0', 'libusb01',
                          'openusb')
            um = __import__('usb.backend', globals(), locals(),
                            candidates, 0)
            for c in candidates:
                try:
                    m = getattr(um, c)
                except AttributeError:
                    continue
                backend = m.get_backend()
                if backend is not None:
                    break
            else:
                raise ValueError('No backend available')
            vp = (vendor, product)
            if nocache or (vp not in cls.UsbDevices):
                # not freed until Python runtime completion
                # enumerate_devices returns a generator, so back up the
                # generated device into a list. To save memory, we only
                # back up the supported devices
                devs = set()
                vpdict = {}
                vpdict.setdefault(vendor, [])
                vpdict[vendor].append(product)
                for dev in backend.enumerate_devices():
                    device = usb.core.Device(dev, backend)
                    vendor = device.idVendor
                    product = device.idProduct
                    if vendor in vpdict:
                        products = vpdict[vendor]
                        if products and (product not in products):
                            continue
                        devs.add(device)
                cls.UsbDevices[vp] = devs
            return cls.UsbDevices[vp]
        finally:
            cls.Lock.release()

    @staticmethod
    def parse_url(urlstr, devclass, scheme, vdict, pdict, default_vendor):
        urlparts = urlsplit(urlstr)
        if scheme != urlparts.scheme:
            raise UsbToolsError("Invalid URL: %s" % urlstr)
        plcomps = urlparts.netloc.split(':') + [''] * 2
        try:
            plcomps[0] = vdict.get(plcomps[0], plcomps[0])
            if plcomps[0]:
                vendor = to_int(plcomps[0])
            else:
                vendor = None
            product_ids = pdict.get(vendor, None)
            if not product_ids:
                product_ids = pdict[default_vendor]
            plcomps[1] = product_ids.get(plcomps[1], plcomps[1])
            if plcomps[1]:
                product = to_int(plcomps[1])
            else:
                product = None
            if not urlparts.path:
                raise UsbToolsError('URL string is missing device port')
            path = urlparts.path.strip('/')
            if path == '?' or (not path and urlstr.endswith('?')):
                show_devices = True
            else:
                interface = to_int(path)
                show_devices = False
        except (IndexError, ValueError):
            raise UsbToolsError('Invalid device URL: %s' % urlstr)
        sernum = None
        idx = 0
        if plcomps[2]:
            try:
                idx = to_int(plcomps[2])
                if idx > 255:
                    idx = 0
                    raise ValueError
                if idx:
                    idx -= 1
            except ValueError:
                sernum = plcomps[2]
        if not vendor or not product or sernum or idx:
            # Need to enumerate USB devices to find a matching device
            vendors = vendor and [vendor] or \
                set(vdict.values())
            vps = set()
            for v in vendors:
                products = pdict.get(v, [])
                for p in products:
                    vps.add((v, products[p]))
            devices = devclass.find_all(vps)
            candidates = []
            if sernum:
                if sernum not in [dev[2] for dev in devices]:
                    raise UsbToolsError("No USB device with S/N %s" % sernum)
                for v, p, s, i, d in devices:
                    if s != sernum:
                        continue
                    if vendor and vendor != v:
                        continue
                    if product and product != p:
                        continue
                    candidates.append((v, p, s, i, d))
            else:
                for v, p, s, i, d in devices:
                    if vendor and vendor != v:
                        continue
                    if product and product != p:
                        continue
                    candidates.append((v, p, s, i, d))
                if not show_devices:
                    try:
                        vendor, product, ifport, ifcount, description = \
                            candidates[idx]
                    except IndexError:
                        raise UsbToolsError("No USB device #%d" % idx)
        if show_devices:
            UsbTools.show_devices(scheme, vdict, pdict, candidates)
            raise SystemExit(candidates and
                             'Please specify the USB device' or
                             'No USB-Serial device has been detected')
        if vendor not in pdict:
            raise UsbToolsError('Vendor ID 0x%04x not supported' % vendor)
        if product not in pdict[vendor].values():
            raise UsbToolsError('Product ID 0x%04x not supported' % product)
        return vendor, product, interface, sernum, idx

    @staticmethod
    def show_devices(scheme, vdict, pdict, candidates, out=None):
        """Show supported devices"""
        from string import printable as printablechars
        if not out:
            import sys
            out = sys.stdout
        indices = {}
        interfaces = []
        for (v, p, s, i, d) in candidates:
            ikey = (v, p)
            indices[ikey] = indices.get(ikey, 0) + 1
            # try to find a matching string for the current vendor
            vendors = []
            # fallback if no matching string for the current vendor is found
            vendor = '%04x' % v
            for vc in vdict:
                if vdict[vc] == v:
                    vendors.append(vc)
            if vendors:
                vendors.sort(key=len)
                vendor = vendors[0]
            # try to find a matching string for the current vendor
            # fallback if no matching string for the current product is found
            product = '%04x' % p
            try:
                products = []
                productids = pdict[v]
                for pc in productids:
                    if productids[pc] == p:
                        products.append(pc)
                if products:
                    products.sort(key=len)
                    product = products[0]
            except KeyError:
                pass
            # if the serial number is an ASCII char, use it, or use the index
            # value
            if not s:
                s = ''
            if [c for c in s if c not in printablechars or c == '?']:
                serial = '%d' % indices[ikey]
            else:
                serial = s
            # Now print out the prettiest URL syntax
            for j in range(1, i+1):
                # On most configurations, low interfaces are used for MPSSE,
                # high interfaces are dedicated to UARTs
                interfaces.append((scheme, vendor, product, serial, j, d))
        if interfaces:
            print_("Available interfaces:", file=out)
            for scheme, vendor, product, serial, j, d in interfaces:
                fmt = '  %s://%s/%d  '
                parts = [vendor, product]
                if serial:
                    parts.append(serial)
                desc = d and '  (%s)' % d
                # the description may contain characters that cannot be
                # emitted in the output stream encoding format
                try:
                    print_(fmt % (scheme, ':'.join(parts), j), end='',
                           file=out)
                except Exception:
                    print_(fmt % (scheme, ':'.join([vendor, product, '???']),
                                  j), end='', file=out)
                try:
                    print_(desc or '', file=out)
                except Exception:
                    print_('', file=out)
            print_('', file=out)

    @classmethod
    def get_string(cls, device, strname):
        """Retrieve a string from the USB device, dealing with PyUSB API breaks
        """
        if cls.UsbApi is None:
            import inspect
            args, varargs, varkw, defaults = \
                inspect.getargspec(usb.util.get_string)
            if (len(args) >= 3) and args[1] == 'length':
                cls.UsbApi = 1
            else:
                cls.UsbApi = 2
        if cls.UsbApi == 2:
            return usb.util.get_string(device, strname)
        else:
            return usb.util.get_string(device, 64, strname)
